"""
Module realizes base plugin class
"""

from dataclasses import dataclass
from typing import Awaitable, TypeVar

from luna3.common.exceptions import LunaApiException
from luna3.common.requests import makeRequest
from sanic.views import HTTPMethodView
from stringcase import camelcase, spinalcase

from ..base.plugins_meta.base_plugins import BaseBackgroundHandler, BaseBackgroundPlugin

try:
    from crutches_on_wheels.cow.errors.errors import Error
    from crutches_on_wheels.cow.errors.exception import VLException
except ImportError:
    from warnings import warn

    warn("Failed to import required modules from `luna-api`")


T_LUNA_APPLICATION = TypeVar("T_LUNA_APPLICATION")
PREVIEW_CHUNK_SIZE = 1024**3


@dataclass
class ProxyPluginRoute:
    """Proxy plugin route description class"""

    # route
    route: str
    # all allowed methods including OPTIONS
    allowedMethods: list[str] = None
    # token permissions | not applicable if not specified
    tokenPermission: str | None = None
    # whether to include service api version to final route
    includeServiceApiVersion: bool = True
    # whether to enable streaming for GET request
    streamGetRequest: bool = False


class BaseLunaApiProxyPlugin(BaseBackgroundPlugin):
    """
    Base api proxy plugin
    """

    async def close(self):
        """
        By default, api proxy plugin don't need any specific actions on application shutdown
        But it is possible to override default `close` method to make some actions
        """

    async def start(self):
        """
        By default, api proxy plugin don't need any specific actions on application start
        But it is possible to override default `start` method to make some actions
        """

    @property
    def pluginPrefix(self) -> str | None:
        """
        Get plugin routes prefix
        """
        return spinalcase(self.pluginName)

    @property
    def whiteList(self) -> list[str]:
        """
        Get plugin routes white list available without authorization
        """
        return []

    @property
    def checkConnections(self) -> list[Awaitable[bool]]:
        """
        Get check connections for plugin
        """
        return []

    @property
    def routes(self) -> list[ProxyPluginRoute]:
        """
        Map with route-to-token-permissions data
        """
        raise NotImplemented

    @property
    def baseProxyHandler(self) -> HTTPMethodView:
        """Base proxy handler from luna-api"""
        raise NotImplemented

    async def initialize(self):
        """
        Description see :func:`~BaseLunaApiProxyPlugin.initialize`.
        """
        for checkConnectionCoro in self.checkConnections:
            if not (await checkConnectionCoro):
                raise Warning(f"Failed to initialize `{self.pluginName}` proxy plugin")

        prefix = f"plugins/{self.pluginPrefix}"
        defaultAllowedMethods = ("POST", "PUT", "GET", "PATCH", "HEAD", "DELETE", "OPTIONS")

        for routeDescription in self.routes:
            allowedMethods_ = routeDescription.allowedMethods or defaultAllowedMethods
            streamGetRequest = routeDescription.streamGetRequest
            route = routeDescription.route[1:] if routeDescription.route.startswith("/") else routeDescription.route
            whiteListRoute = route in self.whiteList

            def getHandler():
                includeServiceApiVersion = routeDescription.includeServiceApiVersion

                class Handler(self.baseProxyHandler, BaseBackgroundHandler):
                    allowedMethods = allowedMethods_

                    def checkTokenPermissions(self) -> None:
                        """
                        Description see :func:`~BaseRequestHandler.checkTokenPermissions`.
                        """
                        if routeDescription.tokenPermission is None:
                            return
                        permissionName = camelcase(routeDescription.tokenPermission)
                        self.checkTokenPermissionsDefault(permissionName)

                    def prepareUrl(self) -> str:
                        """
                        Prepare url to a service
                        Returns:
                            same url with correct api version
                        """
                        uriWithoutApiVersion = self.request.server_path.replace(
                            f"/{self.app.ctx.apiVersion}", "", 1
                        ).replace(f"/{prefix}", "", 1)
                        if includeServiceApiVersion:
                            return (
                                f"{self.serviceAddress.origin}/{self.serviceAddress.apiVersion}{uriWithoutApiVersion}"
                            )
                        return f"{self.serviceAddress.origin}{uriWithoutApiVersion}"

                    async def streamGet(self, **kwargs):
                        """Stream get request"""
                        self.request.streamResponse = True
                        self.checkTokenPermissions()

                        reply = await makeRequest(
                            self.prepareUrl(),
                            method="GET",
                            asyncRequest=True,
                            stream=True,
                            raiseError=True,
                            queryParams=dict(self.request.query_args),
                        )

                        try:
                            async with reply.sendRequest() as resp:
                                headers = dict(resp.headers)
                                headers["Luna-Request-Id"] = self.requestId
                                response = await self.request.respond(headers=headers)
                                async for chunk in resp.iter(PREVIEW_CHUNK_SIZE):
                                    await response.send(chunk)
                        except LunaApiException as exc:
                            raise VLException(Error.PluginError.format(exc), reply.statusCode, False)
                        except Exception as exc:
                            raise VLException(Error.PluginError.format(exc), 500, False)

                if streamGetRequest:
                    Handler.get = Handler.streamGet
                if whiteListRoute:
                    Handler.checkTokenPermissions = lambda x: True

                return Handler

            self.addRoute(
                route=f"/{self.app.apiVersion}/{prefix}/{route}", handler=getHandler(), methods=allowedMethods_
            )
            if whiteListRoute:
                self.app.addPluginRoute2WhiteList(route=route, pluginPrefix=self.pluginPrefix)
