"""
Module import several entities from luna-api for usage by api proxy plugins
If import fails raise en error
Note: required to separate with `BaseLunaApiProxyPlugin` to prevent circular import
"""

from numbers import Real
from typing import Any, Awaitable, Dict, NamedTuple, Optional, Union

from multidict import CIMultiDictProxy

try:
    from app.handlers.base_handler import BaseProxyHandler as _BaseProxyHandler
    from configs.configs.configs.settings.classes import (
        ServiceAddressSettings as _ServiceAddressSettings,
        ServiceTimeoutsSettings as _ServiceTimeoutsSettings,
    )
    from crutches_on_wheels.cow.utils.check_connection import _checkServiceConnection
except ImportError:
    raise ImportError("Failed to import required modules from `luna-api`")

__all__ = [
    "BaseProxyHandler",
    "ProxyRequest",
    "ServiceAddressSettings",
    "ServiceTimeoutsSettings",
    "checkServiceConnection",
]


class ProxyRequest(NamedTuple):
    """Proxy request"""

    body: Any  # body for sending to a proxed service
    headers: Union[Dict[str, str], CIMultiDictProxy[str]]  # headers for sending to a proxed service
    query: Dict[str, Union[None, str]]  # query for sending to a proxed service


class ServiceAddressSettings(_ServiceAddressSettings):
    """Service address setting"""

    origin: str  # service origin, for example: `http://127.0.0.1:12345`
    apiVersion: int  # service api version, for example: `1`


class ServiceTimeoutsSettings(_ServiceTimeoutsSettings):
    """Service timeouts settings"""

    connectTimeout: float  # connect to service timeout
    totalTimeout: float  # total timeout
    sockConnectTimeout: float  # socket connect timeout
    sockReadTimeout: float  # socker read timeout


class BaseProxyHandler(_BaseProxyHandler):
    """Base proxy handler"""

    def checkTokenPermissionsDefault(self, objectName: str) -> None:
        """
        Check token permissions for request using default method to permission map
        """
        return super().checkTokenPermissionsDefault(objectName)

    def prepareUrl(self) -> str:
        """
        Prepare url to a service
        Returns:
            same url with correct api version
        """
        return super().prepareUrl()

    @property
    def app(self) -> "ApiApp":
        """
        Get running app
        Returns:
            app
        """
        return super().app

    @property
    def serviceAddress(self) -> ServiceAddressSettings:
        """
        Get a proxed service address
        Returns:
            a proxed service address
        """
        return super().serviceAddress

    @property
    def serviceTimeouts(self) -> ServiceTimeoutsSettings:
        """
        Get service timeouts
        Returns:
            service timeouts
        """
        return super().serviceTimeouts

    async def prepareRequestPost(self) -> ProxyRequest:
        """
        Prepare proxy request for method `post`.
        """
        return await super().prepareRequestPost()

    async def prepareRequestPut(self) -> ProxyRequest:
        """
        Prepare proxy request for method `put`.
        """
        return await super().prepareRequestPut()

    async def prepareRequestGet(self) -> ProxyRequest:
        """
        Prepare proxy request for method `get`.
        """
        return await super().prepareRequestGet()

    async def prepareRequestPatch(self) -> ProxyRequest:
        """
        Prepare proxy request for method `patch`.
        """
        return await super().prepareRequestPatch()

    async def prepareRequestDelete(self) -> ProxyRequest:
        """
        Prepare proxy request for method `delete`.
        """
        return await super().prepareRequestDelete()

    async def prepareRequestHead(self) -> ProxyRequest:
        """
        Prepare proxy request for method `heada`.
        """
        return await super().prepareRequestHead()


def checkServiceConnection(
    logger: "Logger",
    client: "luna3.common.base_api.BaseAPI",
    service: str,
    closeSession: Optional[bool] = None,
    asyncCheck: bool = False,
    totalTimeout: Optional[Real] = None,
    connectTimeout: Optional[Real] = None,
    sockReadTimeout: Optional[Real] = None,
    sockConnectTimeout: Optional[Real] = None,
) -> Union[bool, Awaitable[bool]]:
    """
    Check connection to any initialised BaseAPI.

    Args:
        logger: logger to log
        client: client to check
        service: service name
        closeSession: whether to close session (True - yes, False - no, None - default)
        asyncCheck: make synchronous or asynchronous check
        totalTimeout: total request timeout
        connectTimeout: socket connect timeout or connection wait from pool timeout
        sockReadTimeout: timeout for reading a chunk from socket
        sockConnectTimeout: socket connect timeout

    Returns:
        True, if check was passed successfully, else False.
    """
    return _checkServiceConnection(
        logger=logger,
        client=client,
        service=service,
        closeSession=closeSession,
        asyncCheck=asyncCheck,
        totalTimeout=totalTimeout,
        connectTimeout=connectTimeout,
        sockReadTimeout=sockReadTimeout,
        sockConnectTimeout=sockConnectTimeout,
    )
